/*
 * memtest version 4.14
 *
 * Very simple but very effective user-space memory tester.
 * Originally by Simon Kirby <sim@stormix.com> <sim@neato.org>
 * Version 2 by Charles Cazabon <memtest@discworld.dyndns.org>
 * Version 3 not publicly released.
 * Version 4 rewrite:
 * Copyright (C) 2004 Charles Cazabon <memtest@discworld.dyndns.org>
 * Copyright (C) 2004 Tony Scaminaci (Version 4.04M port to Macintosh)
 * Copyright (C) 2005 Tony Scaminaci (Versions 4.05M, 4.06M)
 * Licensed under the terms of the GNU General Public License version 2 (only).
 * See the file COPYING for details.
 *
 */

#define __version__ "4.14"
#define EXIT_FAIL_NONSTARTER    0x01
#define EXIT_FAIL_ADDRESSLINES  0x02
#define EXIT_FAIL_OTHERTEST     0x04

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <unistd.h>
#include <errno.h>

// MacOS X includes (Tony Scaminaci 8/2004)

#include <sys/sysctl.h>
#include <sys/file.h>
#include <mach/mach.h>
#include <mach/bootstrap.h>
#include <mach/host_info.h>
#include <mach/mach_error.h>
#include <mach/mach_types.h>
#include <mach/message.h>
#include <mach/vm_region.h>
#include <mach/vm_map.h>
#include <mach/vm_types.h>
#include <mach/vm_prot.h>
#include <mach/shared_memory_server.h>

#include "memtest.h"

// #include <IOKit/IOTypes.h>
// #include <IOKit/IOKitLib.h>

// MacOS X global variables

FILE		*logfile = NULL;
char		*temp_memory;

// MacOS X function prototypes

unsigned long long GetFreeMem(void);

struct test tests[] = {
	{ "Random Value", test_random_value },
	{ "Compare XOR", test_xor_comparison },
	{ "Compare SUB", test_sub_comparison },
	{ "Compare MUL", test_mul_comparison },
	{ "Compare DIV",test_div_comparison },
	{ "Compare OR", test_or_comparison },
	{ "Compare AND", test_and_comparison },
	{ "Sequential Increment", test_seqinc_comparison },
	{ "Solid Bits", test_solidbits_comparison },
	{ "Block Sequential", test_blockseq_comparison },
	{ "Checkerboard", test_checkerboard_comparison },
	{ "Bit Spread", test_bitspread_comparison },
	{ "Bit Flip", test_bitflip_comparison },
	{ "Walking Ones", test_walkbits1_comparison },
	{ "Walking Zeroes", test_walkbits0_comparison },
	{ NULL, NULL }
};

void check_posix_system(void)
  {
    if (sysconf(_SC_VERSION) < 198808L)
	  {
        fprintf(stderr, "WARNING: POSIX version is older than 199908.\n");
        fprintf(stderr, "_SC_VERSION is %lu\n", sysconf(_SC_VERSION));
		if (logfile != NULL)
		  {
			fprintf(logfile, "WARNING: POSIX version is older than 199908.\n");
			fprintf(logfile, "_SC_VERSION is %lu\n", sysconf(_SC_VERSION));
		  }
      }
	else
	  {
		printf("POSIX version %lu\n", sysconf(_SC_VERSION));
		if (logfile != NULL)
		  fprintf(logfile, "POSIX version %lu\n", sysconf(_SC_VERSION));
	  }
  }

// For MacOS X, get amount of free physical RAM at the instant GetFreeMem is called
// Added by Tony Scaminaci 8/15/2004
// Modified function/variable lengths to unsigned long long (64-bit) as required
// to support installed memory sizes > 4 GB. Added by Tony Scaminaci 5/13/2005 

unsigned long long GetFreeMem()
{
  vm_statistics_data_t		vm_stat;
  mach_port_t				host_priv_port;
  mach_msg_type_number_t	host_count;
  kern_return_t				kern_error;
  unsigned long long		FreeMem;		// Free real (physical) memory

// Get total system-wide memory usage structure

  host_priv_port = mach_host_self();	// Get host machine information
  host_count = sizeof(vm_stat)/sizeof(integer_t);
  kern_error = host_statistics(host_priv_port, HOST_VM_INFO, (host_info_t)&vm_stat, &host_count);
  if (kern_error != KERN_SUCCESS)
	{
	  mach_error("host_info", kern_error);
	  exit(EXIT_FAILURE);
	}
  FreeMem = ((unsigned long long) vm_stat.free_count) * vm_page_size;	// Calculate total free memory in bytes
  FreeMem &= 0xFFFFFFFFFFFFFFC0ull;						// Align byte count to 64-byte boundary
  return(FreeMem);
}

// Check current user level. If it's 0, we're in single-user mode, if it's 1, we're in multiuser mode.

int	get_user_level()
  {
	int		Selectors[2] = {CTL_KERN, KERN_SECURELVL};
	int		UserLevel;
	size_t	Length = sizeof(UserLevel);
	
	sysctl(Selectors, 2, &UserLevel, &Length, NULL, (size_t) 0);
	return UserLevel;
  }
  
void store_dimm_data(char data)
  {
	int	index = 0;
	
	printf("Character: %c", data);
	temp_memory[index++] = data;  
  }

void disableMallocDebug(char *const argv[])
  {
    if (getenv("MallocLogFile") == NULL)
	  {
		if (putenv("MallocLogFile=/dev/null") == -1)
		  {
			perror("ERROR: putenv");
			exit(1);
		  }
		if (execvp(argv[0], argv) == -1)
		  {
			perror("ERROR: execvp");
			exit(1);
		  }
	  }
  }
  
// Get CPU information
  
int Get_CPUInfo(void)
 {
	int		hasAltiVec = 0, hasSSE3 = 0, NoOfCPUs = 0, ByteOrder = 0;
	int		CPUType = 0;	// Initialize to Intel CPU
	size_t	Length = sizeof(int);
	
	sysctlbyname("hw.ncpu", &NoOfCPUs, &Length, NULL, (size_t) 0);
	sysctlbyname("hw.byteorder", &ByteOrder, &Length, NULL, (size_t) 0);
	if (ByteOrder == 4321)	// 4321 = Big Endian --> CPU is PPC
	  {
		CPUType = 1;	// CPU is PowerPC
		sysctlbyname("hw.optional.altivec", &hasAltiVec, &Length, NULL, (size_t) 0);
		return (hasAltiVec + (NoOfCPUs << 8) + (CPUType << 16));
	  }
	else				// 1234 = Little Endian --> CPU is Intel
	  {
		sysctlbyname("hw.optional.sse3", &hasSSE3, &Length, NULL, (size_t) 0);
		return (hasSSE3 + (NoOfCPUs << 8) + (CPUType << 16));
	  }
}

int main(int argc, char *const argv[]) {
    int loop, i;
	int loops;
    size_t halflen, count, bufsize;
    ptrdiff_t pagesizemask;
    void volatile *buf, *aligned;
    ulv *bufa, *bufb;
    int do_mlock = 1, done_mem = 0;
	int exit_code = 0;
	int	NoArgFlag = 0;					// Set if no arguments are passed to application
	unsigned long MBAvail;				// Physical memory available for testing (MB)
	unsigned long wantmb;				// Physical memory requested for lock (MB)
	unsigned long long MemAvail;		// Physical memory available for testing (bytes)
	unsigned long long wantbytes;		// Physical memory requested for lock (bytes)
	unsigned long long wantbytes_orig;	// Copy of physical memory requested for lock (bytes)
	unsigned long long testbytes;		// Test size for single-user mode
	int	user_level;						// Single-user or multiuser mode
	int CPUInfo;						// CPU and optional vector accelerator information
//	int errcode;

//	Disable debugging messages from the malloc library. These are generated whenever malloc
//	cannot allocate the buffer size as requested. As the requests are repeatedly reduced by
//	one page size, the warnings will normally scroll continuously on the screen. Instead,
//	route them to /dev/null.

    disableMallocDebug(argv);

//  IOKit test code
//	io_service_t	device;
//	io_name_t		devName;
//	io_string_t		pathName;

//	IORegistryEntryGetName(device, devName);
//	printf("Devices name = %s\n", devName);
//	IORegistryEntryGetPath(device, kIOServicePlane, pathName);
//	printf(Devices path in IOService plane = %s\n, pathName);
//	IORegistryEntryGetPath(device, kIOUSBPlane, pathName);
//	printf(Devices path in IOUSB plane = %s\n, pathName);
//  End of IOKit test code

//	temp_memory = malloc(1024);
//	system("ioreg -p IODeviceTree -l | grep memory@0 -A 11");
//	free(temp_memory);

    printf("\nMemtest version " __version__ " (%lu-bit)\n", sizeof(size_t) << 3);
//	printf("Maximum value of unsigned long: %lu\n", ULONG_MAX);
//	printf("Size of size_t = %zd\n", sizeof(size_t));
    printf("Copyright (C) 2004 Charles Cazabon\n");
	printf("Copyright (C) 2004, 2005 Tony Scaminaci (Macintosh port)\n");
    printf("Licensed under the GNU General Public License version 2 only\n\n");
	if (argc < 2)
	  {
		NoArgFlag = 1;
		printf("NOTE: No command-line arguments have been specified\n");
		printf("Using defaults: test all available memory, 1 pass, no logfile\n\n");
	  }
	user_level = get_user_level();
	if (argc == 4 && (!strcmp (argv[3], "-l") || !strcmp (argv[3], "-L")))
	  {
		if (!user_level)								// If we're in single-user mode,
		  system("/sbin/mount -uw /");					// make / file system writable
		logfile = fopen ("memtest.log", "w");
		if (logfile == NULL)
		  printf("Couldn't open logfile... continuing anyway\n");
		else
		  {
			printf("Logging to file 'memtest.log' in current working directory\n\n");
//			errcode = flock((int)logfile, LOCK_EX | LOCK_NB);
//			printf("FLock error = %d FLock status = %d\n", errcode, errno);
		  }
	  }
	if (logfile != NULL)
	  {
		fprintf(logfile, "\nMemtest version " __version__ " (%lu-bit)\n", sizeof(size_t) << 3);
		fprintf(logfile, "Copyright (C) 2004 Charles Cazabon\n");
		fprintf(logfile, "Copyright (C) 2004, 2005, 2006 Tony Scaminaci (Macintosh port)\n");
		fprintf(logfile, "Licensed under the GNU General Public License version 2 only\n\n");
	  }
	if (!user_level)
	  printf("MacOS X (Darwin) running in single user mode\n");
	else
	  printf("MacOS X (Darwin) running in multiuser mode\n");
	if (logfile != NULL)
	  {
		if (!user_level)
		  fprintf(logfile, "MacOS X (Darwin) running in single user mode\n");
		else
		  fprintf(logfile, "MacOS X (Darwin) running in multiuser mode\n");
	  }
	check_posix_system();
	printf("Pagesize is %lu\n", (unsigned long) vm_page_size);
	if (logfile != NULL)
	  fprintf(logfile, "Pagesize is %lu\n", (unsigned long) vm_page_size);
    pagesizemask = (ptrdiff_t) ~(vm_page_size - 1);
//	printf("Pagesizemask is 0x%tx\n", pagesizemask);
	CPUInfo = Get_CPUInfo();
	if ((CPUInfo >> 16) & 0x00FF)	// Check for PPC CPU
	  {
		if (CPUInfo & 0x00FF)		// Check for Altivec unit
		  {
			printf("System has %d PPC processor(s) with Altivec\n", ((CPUInfo >> 8) & 0x00FF));
			if (logfile != NULL)
			  {
				fprintf(logfile, "System has %d PPC processor(s) with Altivec\n", ((CPUInfo >> 8) & 0x00FF));
			  }
		  }
		else
		  {
			printf("System has %d PPC processor(s) without Altivec\n", ((CPUInfo >> 8) & 0x00FF));
			if (logfile != NULL)
			  {
				fprintf(logfile, "System has %d PPC processor(s) without Altivec\n", ((CPUInfo >> 8) & 0x00FF));
			  }
		  }
	  }
	else							// CPU is Intel
	  {
		if (CPUInfo & 0x00FF)		// Check for SSE unit
		  {
			printf("System has %d Intel processor(s) with SSE\n", ((CPUInfo >> 8) & 0x00FF));
			if (logfile != NULL)
			  {
				fprintf(logfile, "System has %d Intel processor(s) with SSE\n", ((CPUInfo >> 8) & 0x00FF));
			  }
		  }
		else
		  {
			printf("System has %d Intel processor(s) without SSE\n", ((CPUInfo >> 8) & 0x00FF));
			if (logfile != NULL)
			  {
				fprintf(logfile, "System has %d Intel processor(s) without SSE\n", ((CPUInfo >> 8) & 0x00FF));
			  }
		  }
	  }
	MemAvail = GetFreeMem();			// Get 64-bit length of free memory in bytes
	MBAvail = MemAvail >> 20;			// Calculate free memory in megabytes
	if (NoArgFlag || !strcmp(argv[1], "all") || !strcmp(argv[1], "ALL"))
	  {
	    wantmb = MBAvail;						// "all" option specified so set requested
		wantbytes_orig = wantbytes = MemAvail;	// memory size equal to available memory
	  }
	else								// Else, set requested memory size to user's request
	  {
		wantmb = (unsigned long) strtoul(argv[1], NULL, 0);
		wantbytes_orig = wantbytes = (unsigned long long) (wantmb << 20);
	  }
	printf("Requested memory: %lluMB (%llu bytes)\n", (ull) wantmb, (ull) wantbytes);
	printf("Available memory: %lluMB (%llu bytes)\n", (ull) MBAvail, (ull) MemAvail);
	if (logfile != NULL)
	  {
		fprintf(logfile, "Requested memory: %lluMB (%llu bytes)\n", (ull) wantmb, (ull) wantbytes);
		fprintf(logfile, "Available memory: %lluMB (%llu bytes)\n", (ull) MBAvail, (ull) MemAvail);
	  }
    if (wantbytes < (unsigned long long) vm_page_size)
	  {
        fprintf(stderr, "\nERROR: Memory test size in MB (argument 2) must be a positive integer or 'all'.\n\n");
	        exit(EXIT_FAIL_NONSTARTER);
      }
	if (user_level && wantbytes > MemAvail)			// Limit memory allocation in multiuser modes
	  {
		printf("NOTE: Memory request is too large, reducing to acceptable value...\n");
		if (logfile != NULL)
		  fprintf(logfile, "NOTE: Memory request is too large, reducing to acceptable value...\n");
		wantbytes = MemAvail;						// Guarantee that no paging will occur
	  }
//	In single-user mode, reduce the allocation request by a preset factor and align to 64-byte boundary
	testbytes = (((unsigned long long) (MemAvail * 0.972)) >> 6) << 6;
	if (!user_level && wantbytes > testbytes)		// Further limit memory allocation in single-user mode
	  {
		printf("NOTE: Memory request is too large, reducing to acceptable value...\n");
		if (logfile != NULL)
		  fprintf(logfile, "NOTE: Memory request is too large, reducing to acceptable value...\n");
		wantbytes = testbytes;						// Prevent the Darwin kernel from becoming unresponsive
	  }
//	Finally, limit the allocation request size to the maximum memory that the platform (32 or 64-bit) can address
	if (wantbytes > ULONG_MAX)
	  wantbytes = ULONG_MAX;
    if (argc < 3)
	  {
        loops = 1;
      }
	else
	  {
        loops = strtoul(argv[2], NULL, 0);
		if (loops < 0)
		  loops = 0;
      }	  
    buf = NULL;							// Initialize memory test buffer pointer
    while (!done_mem)
	  {
        while (!buf && wantbytes)		// Attempt to allocate requested number of bytes
		  {
            buf = (void volatile *) malloc((size_t) wantbytes);
            if (!buf)						// If allocation request fails, reduce requested
			  wantbytes -= vm_page_size;	// allocation by one page and try again
          }
        bufsize = (size_t) wantbytes;
        if ((size_t) buf % vm_page_size)	// If buffer is not aligned to a memory page boundary
		  {
            printf("Aligning test buffer to nearest page boundary...\n");
			if (logfile != NULL)
			  fprintf(logfile, "Aligning test buffer to nearest page boundary...\n");
			fflush(stdout);
            aligned = (void volatile *) ((size_t) buf & pagesizemask);
            bufsize -= ((size_t) aligned - (size_t) buf);
          }
		else
		  aligned = buf;				// Test buffer was already aligned to a page boundary
		if (sizeof(size_t) == 8)
		  printf("Allocated memory: %lluMB (%llu bytes) at local address %.16p\n", (ull) bufsize >> 20, (ull) bufsize, aligned);
		else
		   printf("Allocated memory: %lluMB (%llu bytes) at local address %.8p\n", (ull) bufsize >> 20, (ull) bufsize, aligned);
		if (logfile != NULL)
		  {
			if (sizeof(size_t) == 8)
			  fprintf(logfile, "Allocated memory: %lluMB (%llu bytes) at local address %.16p\n", (ull) bufsize >> 20, (ull) bufsize, aligned);
			else
			  fprintf(logfile, "Allocated memory: %lluMB (%llu bytes) at local address %.8p\n", (ull) bufsize >> 20, (ull) bufsize, aligned);
		  }
//	The buffer is now aligned to a memory page boundary so attempt to lock the memory for exclusive use
		printf("Attempting memory lock... ");
		if (logfile != NULL)
		  fprintf(logfile, "Attempting memory lock... ");
		fflush(stdout);
        if (mlock((void *) aligned, bufsize) < 0)
		  {
            switch(errno)
			  {
                case ENOMEM:
                    printf("too many pages requested - reducing and trying again...\n\n");
					if (logfile != NULL)
					  fprintf(logfile, "too many pages requested - reducing and trying again...\n\n");
                    free((void *) buf);
                    buf = NULL;
                    wantbytes -= vm_page_size;
                    break;
                case EPERM:
                    printf("insufficient permissions - running unlocked\n\n");
					if (logfile != NULL)
					  fprintf(logfile, "insufficient permissions - running unlocked\n\n");
                    do_mlock = 0;
					free((void *) buf);
					buf = NULL;
					wantbytes = wantbytes_orig;
                    break;
                default:
                    printf("ERROR: Memory lock failed - reason unknown.\n\n");
					if (logfile != NULL)
					  fprintf(logfile, "ERROR: Memory lock failed - reason unknown.\n\n");
                    do_mlock = 0;
                    done_mem = 1;
             }
		  }
		else
		  {
            printf("locked successfully\n");
			if (logfile != NULL)
			  fprintf(logfile, "locked successfully\n");
			fflush(stdout);
            done_mem = 1;
          }
    }

    if (!do_mlock)
	  {
		fprintf(stderr, "WARNING: Testing with unlocked memory may be slower and less reliable\n\n");
		if (logfile != NULL)
		  fprintf(logfile, "WARNING: Testing with unlocked memory may be slower and less reliable\n\n");
	  }

    halflen = bufsize / 2;
    count = halflen / sizeof(ul);
    bufa = (ulv *) aligned;
    bufb = (ulv *) ((size_t) aligned + halflen);
	printf("Creating test buffers...\n");
	if (sizeof(size_t) == 8)
	  {
		printf("Buffer A: %lluMB (%llu bytes) at local address %.16p\n", (ull) halflen >> 20, (ull) halflen, bufa);
		printf("Buffer B: %lluMB (%llu bytes) at local address %.16p\n\n", (ull) halflen >> 20, (ull) halflen, bufb);
	  }
	else
	  {
		printf("Buffer A: %lluMB (%llu bytes) at local address %.8p\n", (ull) halflen >> 20, (ull) halflen, bufa);
		printf("Buffer B: %lluMB (%llu bytes) at local address %.8p\n\n", (ull) halflen >> 20, (ull) halflen, bufb);
	  }	
	fflush(stdout);
	if (logfile != NULL)
	  {
	  	fprintf(logfile, "Creating test buffers...\n");
		if (sizeof(size_t) == 8)
		  {
			fprintf(logfile, "Buffer A: %lluMB (%llu bytes) at local address %.16p\n", (ull) halflen >> 20, (ull) halflen, bufa);
			fprintf(logfile, "Buffer B: %lluMB (%llu bytes) at local address %.16p\n\n", (ull) halflen >> 20, (ull) halflen, bufb);
		  }
		else
		  {
			fprintf(logfile, "Buffer A: %lluMB (%llu bytes) at local address %.8p\n", (ull) halflen >> 20, (ull) halflen, bufa);
			fprintf(logfile, "Buffer B: %lluMB (%llu bytes) at local address %.8p\n\n", (ull) halflen >> 20, (ull) halflen, bufb);
		  }
		fflush(logfile);
	  }
	if (!loops)
	  {
		printf("NOTE: Test sequences will run continuously until terminated by Ctrl-C...\n\n");
		if (logfile != NULL)
		  fprintf(logfile, "NOTE: Test sequences will run continuously until terminated by Ctrl-C...\n\n");
	  }
	else
	  {
		if (loops == 1)
		  {
			printf("Running %d test sequence...\n\n", loops);
			if (logfile != NULL)
			  fprintf(logfile, "Running %d test sequence...\n\n", loops);
		  }
		else
		  {
			printf("Running %d test sequences...\n\n", loops);
			if (logfile != NULL)
			  fprintf(logfile, "Running %d test sequences...\n\n", loops);
		  }
	  }
    for(loop = 1; ((!loops) ||  loop <= loops); loop++) {
        printf("Test sequence %d", loop);
		if (logfile != NULL)
		  fprintf(logfile, "Test sequence %d", loop);
        if (loops) {
            printf(" of %d", loops);
			if (logfile != NULL)
			  fprintf(logfile, " of %d", loops);
        }
        printf(":\n");
    	printf("  %-20s:    ", "Stuck Address");
		fflush(stdout);
		if (logfile != NULL)
		  {
			fprintf(logfile, ":\n");
			fprintf(logfile, "  %-20s:    ", "Stuck Address");
			fflush(logfile);
		  }
    	if (!test_stuck_address(aligned, bufsize / sizeof(ul)))
		  {
			printf("ok\n");
			if (logfile != NULL)
			  {
				fprintf(logfile, "ok\n");
				fflush(logfile);
			  }
		  }
		else
		  exit_code |= EXIT_FAIL_ADDRESSLINES;		
    	for (i=0;;i++) {
		  if (!tests[i].name) break;
		    printf("  %-20s:    ", tests[i].name);
			if (logfile != NULL)
			  fprintf(logfile, "  %-20s:    ", tests[i].name);
		  if (!tests[i].fp(bufa, bufb, count))
		    {
			  printf("ok\n");
			  if (logfile != NULL)
				fprintf(logfile, "ok\n");
			}
		  else
			exit_code |= EXIT_FAIL_OTHERTEST;
		  fflush(stdout);
		  if (logfile != NULL)
			fflush(logfile);
    	}
    	printf("\n");
		fflush(stdout);
		if (logfile != NULL)
		  {
			fprintf(logfile, "\n");
			fflush(logfile);
		  }
    }
    if (do_mlock) munlock((void *) aligned, bufsize);
	if (!exit_code)
	  {
		printf("All tests passed.\n\n");
		if (logfile != NULL)
		  {
			fprintf(logfile, "All tests passed.\n");
			fclose(logfile);
		  }
	  }
	else
	  {
		if (exit_code & EXIT_FAIL_ADDRESSLINES)
		  {
			printf("*** Address Test Failed ***  One or more DIMM address lines are non-functional.\n");
			if (logfile != NULL)
			  fprintf(logfile, "*** Address Test Failed ***  One or more DIMM address lines are non-functional.\n");
		  }
		if (exit_code & EXIT_FAIL_OTHERTEST)
		  {
			printf("*** Memory Test Failed ***  Please check transcript for details.\n");
			if (logfile != NULL)
			  fprintf(logfile, "*** Memory Test Failed ***  Please check logfile for details.\n");
		  }
		if (logfile != NULL)
		  fclose(logfile);
	  }
	fflush(stdout);
//	if (!user_level)							// If we're in single-user mode,
//	  system("/sbin/mount -urf /");				// return file system to read-only
    exit(0);
}
