/*
 * Very simple but very effective user-space memory tester.
 * Originally by Simon Kirby <sim@stormix.com> <sim@neato.org>
 * Version 2 by Charles Cazabon <memtest@discworld.dyndns.org>
 * Version 3 not publicly released.
 * Version 4 rewrite:
 * Copyright (C) 2004 Charles Cazabon <memtest@discworld.dyndns.org>
 * Portions Copyright (C) 2005 Tony Scaminaci
 * Licensed under the terms of the GNU General Public License version 2 (only).
 * See the file COPYING for details.
 *
 * This file contains the functions for the actual tests, called from the
 * main routine in memtest.c.  See other comments in that file.
 *
 */

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>

#include "memtest.h"

// Global variables

extern FILE	*logfile;

char progress[] = "-\\|/";
#define PROGRESSLEN 4
#define PROGRESSOFTEN 600000

/* Function definitions. */

int compare_regions(ulv *bufa, ulv *bufb, size_t count)
{
    int r = 0;
    size_t i;

    for (i = 0; i < count; i++)
	{
        if (bufa[i] != bufb[i])
		{
            fprintf(stderr, "FAILURE: 0x%08lx != 0x%08lx at offset 0x%08lx.\n", (ul) bufa[i], (ul) bufb[i], (ul) i);
			if (logfile != NULL)
			  {
				fprintf(logfile, "FAILURE: 0x%08lx != 0x%08lx at offset 0x%08lx.\n", (ul) bufa[i], (ul) bufb[i], (ul) i);
				fflush(logfile);
			  }
            r = -1;
        }
    }
    return r;
}

int test_stuck_address(ulv *buf, size_t count)
{
    unsigned int j;
    size_t i;

	printf("             ");
    fflush(stdout);
    for (j = 0; j < 16; j++)
	{
		printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
        printf("setting %2u of 16", j+1);
        fflush(stdout);
        for (i = 0; i < count; i++)
		{
            buf[i] = ((j + i) % 2) == 0 ? (ul) &buf[i] : ~((ul) &buf[i]);
        }
		printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
        printf("testing %2u of 16", j+1);
		fflush(stdout);
        for (i = 0; i < count; i++)
		{
            if (buf[i] != (((j + i) % 2) == 0 ? (ul) &buf[i] : ~((ul) &buf[i])))
			{
                fprintf(stderr, "FAILURE: possible bad address line at offset 0x%08lx.\n", (ul) i);
                printf("Skipping to next test...\n");
				fflush(stdout);
				if (logfile != NULL)
				  {
					fprintf(logfile, "FAILURE: possible bad address line at offset 0x%08lx.\n", (ul) i);
					fprintf(logfile, "Skipping to next test...\n");
					fflush(logfile);
				  }
                return -1;
            }
        }
    }
    printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b                \b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
    fflush(stdout);
    return 0;
}

int test_random_value(ulv *bufa, ulv *bufb, size_t count)
{
    ul j = 0;
    size_t i;

	printf("\b\b");
    fflush(stdout);
    for (i = 0; i < count; i++)
	{
        bufa[i] = bufb[i] = rand_ul();
		if (!(i % PROGRESSOFTEN))
		{
			putchar('\b');
			putchar(progress[++j % PROGRESSLEN]);
            fflush(stdout);
		}
    }
	printf("\b \b");
    fflush(stdout);
    return compare_regions(bufa, bufb, count);
}

int test_xor_comparison(ulv *bufa, ulv *bufb, size_t count)
{
    size_t i;
    ul q = rand_ul();

    for (i = 0; i < count; i++)
	{
        bufa[i] ^= q;
        bufb[i] ^= q;
    }
	printf("\b\b\b   \b\b\b");
    fflush(stdout);
    return compare_regions(bufa, bufb, count);
}

int test_sub_comparison(ulv *bufa, ulv *bufb, size_t count)
{
    size_t i;
    ul q = rand_ul();

    for (i = 0; i < count; i++)
	{
        bufa[i] -= q;
        bufb[i] -= q;
    }
	printf("\b\b\b   \b\b\b");
    fflush(stdout);
    return compare_regions(bufa, bufb, count);
}

int test_mul_comparison(ulv *bufa, ulv *bufb, size_t count)
{
    size_t i;
    ul q = rand_ul();

    for (i = 0; i < count; i++)
	{
        bufa[i] *= q;
        bufb[i] *= q;
    }
	printf("\b\b\b   \b\b\b");
    fflush(stdout);
    return compare_regions(bufa, bufb, count);
}

int test_div_comparison(ulv *bufa, ulv *bufb, size_t count)
{
    size_t i;
    ul q = rand_ul();

    for (i = 0; i < count; i++)
	{
        if (!q)
		{
            q++;
        }
        bufa[i] /= q;
        bufb[i] /= q;
    }
	printf("\b\b\b   \b\b\b");
    fflush(stdout);
    return compare_regions(bufa, bufb, count);
}

int test_or_comparison(ulv *bufa, ulv *bufb, size_t count)
{
    size_t i;
    ul q = rand_ul();

    for (i = 0; i < count; i++)
	{
        bufa[i] |= q;
        bufb[i] |= q;
    }
	printf("\b\b\b   \b\b\b");
    fflush(stdout);
    return compare_regions(bufa, bufb, count);
}

int test_and_comparison(ulv *bufa, ulv *bufb, size_t count)
{
    size_t i;
    ul q = rand_ul();

    for (i = 0; i < count; i++)
	{
        bufa[i] &= q;
        bufb[i] &= q;
    }
	printf("\b\b\b   \b\b\b");
    fflush(stdout);
    return compare_regions(bufa, bufb, count);
}

int test_seqinc_comparison(ulv *bufa, ulv *bufb, size_t count)
{
    size_t i;
    ul q = rand_ul();

    for (i = 0; i < count; i++)
	{
        bufa[i] = bufb[i] = (i + q);
    }
	printf("\b\b\b   \b\b\b");
    fflush(stdout);
    return compare_regions(bufa, bufb, count);
}

int test_solidbits_comparison(ulv *bufa, ulv *bufb, size_t count)
{
    unsigned int j;
    ul q;
    size_t i;

	printf("             ");
    fflush(stdout);
    for (j = 0; j < 64; j++)
	{
	    printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
        q = (j % 2) == 0 ? UL_ONEBITS : 0;
        printf("setting %2u of 64", j+1);
        fflush(stdout);
        for (i = 0; i < count; i++)
		{
            bufa[i] = bufb[i] = (i % 2) == 0 ? q : ~q;
        }
        printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
        printf("testing %2u of 64", j+1);
        fflush(stdout);
        if (compare_regions(bufa, bufb, count))
		{
            return -1;
        }
    }
    printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b                \b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
    fflush(stdout);
    return 0;
}

int test_checkerboard_comparison(ulv *bufa, ulv *bufb, size_t count)
{
    unsigned int j;
    ul q;
    size_t i;

	printf("             ");
    fflush(stdout);
    for (j = 0; j < 64; j++)
	{
	    printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
        q = (j % 2) == 0 ? CHECKERBOARD1 : CHECKERBOARD2;
        printf("setting %2u of 64", j+1);
        fflush(stdout);
        for (i = 0; i < count; i++)
		{
            bufa[i] = bufb[i] = (i % 2) == 0 ? q : ~q;
        }
        printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
        printf("testing %2u of 64", j+1);
        fflush(stdout);
        if (compare_regions(bufa, bufb, count))
		{
            return -1;
        }
    }
    printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b                \b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
    fflush(stdout);
    return 0;
}

int test_blockseq_comparison(ulv *bufa, ulv *bufb, size_t count)
{
    unsigned int j;
    size_t i;

	printf("               ");
    fflush(stdout);
    for (j = 0; j < 256; j++)
	{
	    printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
        printf("setting %3u of 256", j+1);
        fflush(stdout);
        for (i = 0; i < count; i++)
		{
            bufa[i] = bufb[i] = (ul) UL_BYTE(j);
        }
        printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
        printf("testing %3u of 256", j+1);
        fflush(stdout);
        if (compare_regions(bufa, bufb, count))
		{
            return -1;
        }
    }
    printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b                    \b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
    fflush(stdout);
    return 0;
}

int test_walkbits0_comparison(ulv *bufa, ulv *bufb, size_t count)
{
    unsigned int j;
    size_t i;

	printf("             ");
    fflush(stdout);
    for (j = 0; j < 64; j++)
	{
	    printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
        printf("setting %2u of 64", j+1);
        fflush(stdout);
        for (i = 0; i < count; i++)
		{
            if (j < UL_LEN) 
			{ /* Walk it up. */
                bufa[i] = bufb[i] = 0x00000001 << j;
            } else
			{ /* Walk it back down. */
                bufa[i] = bufb[i] = 0x00000001 << (UL_LEN * 2 - j - 1);
            }
        }
        printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
        printf("testing %2u of 64", j+1);
        fflush(stdout);
        if (compare_regions(bufa, bufb, count))
		{
            return -1;
        }
    }
    printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b                \b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
    fflush(stdout);
    return 0;
}

int test_walkbits1_comparison(ulv *bufa, ulv *bufb, size_t count)
{
    unsigned int j;
    size_t i;

	printf("               ");
    fflush(stdout);
    for (j = 0; j < UL_LEN * 2; j++)
	{
	    printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
		printf("setting %3u of %3u", j+1, UL_LEN * 2);
        fflush(stdout);
        for (i = 0; i < count; i++)
		{
            if (j < UL_LEN)
			{ /* Walk it up. */
                bufa[i] = bufb[i] = UL_ONEBITS ^ (0x00000001 << j);
            } else
			{ /* Walk it back down. */
                bufa[i] = bufb[i] = UL_ONEBITS ^ (0x00000001 << (UL_LEN * 2 - j - 1));
            }
        }
        printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
        printf("testing %3u of %3u", j+1, UL_LEN * 2);
        fflush(stdout);
        if (compare_regions(bufa, bufb, count))
		{
            return -1;
        }
    }
    printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b                  \b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
    fflush(stdout);
    return 0;
}

int test_bitspread_comparison(ulv *bufa, ulv *bufb, size_t count)
{
    unsigned int j;
    size_t i;

	printf("               ");
    fflush(stdout);
    for (j = 0; j < UL_LEN * 2; j++)
	{
	    printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
        printf("setting %3u of %3u", j+1, UL_LEN * 2);
        fflush(stdout);
        for (i = 0; i < count; i++)
		{
            if (j < UL_LEN)
			{ /* Walk it up. */
                bufa[i] = bufb[i] = (i % 2 == 0)
                    ? (0x00000001 << j) | (0x00000001 << (j + 2))
                    : UL_ONEBITS ^ ((0x00000001 << j)
                                    | (0x00000001 << (j + 2)));
            } else
			{ /* Walk it back down. */
                bufa[i] = bufb[i] = (i % 2 == 0)
                    ? (0x00000001 << (UL_LEN * 2 - 1 - j)) | (0x00000001 << (UL_LEN * 2 + 1 - j))
                    : UL_ONEBITS ^ (0x00000001 << (UL_LEN * 2 - 1 - j)
                                    | (0x00000001 << (UL_LEN * 2 + 1 - j)));
            }
        }
        printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
        printf("testing %3u of %3u", j+1, UL_LEN * 2);
        fflush(stdout);
        if (compare_regions(bufa, bufb, count))
		{
            return -1;
        }
    }
    printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b                  \b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
    fflush(stdout);
    return 0;
}

int test_bitflip_comparison(ulv *bufa, ulv *bufb, size_t count)
{
    unsigned int j, k;
    ul q;
    size_t i;

	printf("               ");
    fflush(stdout);
    for (k = 0; k < UL_LEN; k++)
	{
        q = 0x00000001 << k;
        for (j = 0; j < 8; j++)
		{
    	    printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
            q = ~q;
            printf("setting %3u of %3u", k * 8 + j + 1, UL_LEN * 8);
            fflush(stdout);
            for (i = 0; i < count; i++) {
                bufa[i] = bufb[i] = (i % 2) == 0 ? q : ~q;
            }
            printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
            printf("testing %3u of %3u", k * 8 + j + 1, UL_LEN * 8);
            fflush(stdout);
            if (compare_regions(bufa, bufb, count))
			{
                return -1;
            }
        }
    }
    printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b                  \b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
    fflush(stdout);
    return 0;
}
